package src.rsa;

import java.math.BigInteger;
import java.security.SecureRandom;

public class PrimeGenerator {
    private static final SecureRandom random = new SecureRandom();

    // 1️⃣ Implementera Rabin-Miller primtalstest
    public static boolean isPrime(BigInteger n, int k) {
        if (n.equals(BigInteger.ONE) || n.mod(BigInteger.TWO).equals(BigInteger.ZERO)) {
            return false; // 1 eller jämna tal är inte primtal (förutom 2)
        }
    
        // Skriv n-1 som 2^r * s, där s är udda
        BigInteger s = n.subtract(BigInteger.ONE);
        int r = 0;
        while (s.mod(BigInteger.TWO).equals(BigInteger.ZERO)) {
            s = s.divide(BigInteger.TWO);
            r++;
        }
    
        // Kör testet k gånger med slumpmässiga baser
        for (int i = 0; i < k; i++) {
            BigInteger a = uniformRandom(BigInteger.TWO, n.subtract(BigInteger.ONE)); // Slumpmässig bas
            BigInteger x = a.modPow(s, n);
    
            if (x.equals(BigInteger.ONE) || x.equals(n.subtract(BigInteger.ONE))) {
                continue; // Möjligt primtal
            }
    
            boolean isComposite = true;
            for (int j = 0; j < r - 1; j++) {
                x = x.modPow(BigInteger.TWO, n);
                if (x.equals(n.subtract(BigInteger.ONE))) {
                    isComposite = false;
                    break;
                }
            }
    
            if (isComposite) {
                return false; // n är sammansatt
            }
        }
    
        return true; // n är möjligen primtal
    }
    
    // Hjälpmetod för att generera ett slumpmässigt BigInteger mellan min och max
    private static BigInteger uniformRandom(BigInteger min, BigInteger max) {
        BigInteger result;
        do {
            result = new BigInteger(max.bitLength(), random);
        } while (result.compareTo(min) < 0 || result.compareTo(max) > 0);
        return result;
    }

    // 2️⃣ Generera ett primtal med angiven bitstorlek
    public static BigInteger generatePrime(int bitLength) {
        BigInteger prime;
        do {
            prime = new BigInteger(bitLength, random).setBit(0); // Se till att det är udda
        } while (!isPrime(prime, 20)); // Testa med 20 iterationer av Rabin-Miller
        return prime;
    }

    // 3️⃣ Generera 100 primtal av viss storlek och mät tiden
    public static void benchmarkPrimes(int bitLength) {
        long startTime = System.nanoTime();
    
        for (int i = 0; i < 100; i++) {
            generatePrime(bitLength);
        }
    
        long endTime = System.nanoTime();
        double elapsedTime = (endTime - startTime) / 1_000_000_000.0; // Konvertera till sekunder
        System.out.println("Bitlängd: " + bitLength + " | Tid: " + elapsedTime + " sekunder");
    }

    public static void main(String[] args) {
        System.out.println("=== Benchmark: Primtalsgenerering ===");
        System.out.println("| Bitlängd | Antal | Tid (ms) |");
        System.out.println("|----------|-------|----------|");
    
        int[] bitSizes = {512, 1024, 2048, 4096};  // Bitstorlekar att testa
        for (int bitSize : bitSizes) {
            long startTime = System.currentTimeMillis();  // Starta tidtagning
    
            for (int i = 0; i < 100; i++) {
                generatePrime(bitSize);  // Generera primtal
            }
    
            long endTime = System.currentTimeMillis();  // Sluta tidtagning
            long elapsedTime = endTime - startTime;  // Total tid i ms
    
            // Skriv ut tabellrad
            System.out.printf("| %8d | %4d  | %8d |\n", bitSize, 100, elapsedTime);
        }
    
        System.out.println("\nBenchmark färdig! 🎉");
    }
}
